import torch
import torch.nn as nn
import torch.optim as optim

import torch.nn.functional as F
import torchaudio
import tqdm

import argparse

import math

import os

from torch.utils.data import random_split
from torch.utils.data import DataLoader

import AudioUtils
from data import AudioDataset
from generator import SISUGenerator
from discriminator import SISUDiscriminator

import librosa

def mfcc_loss(y_true, y_pred, sr):
    # 1. Ensure sr is a NumPy scalar (not a Tensor)
    if isinstance(sr, torch.Tensor):  # Check if it's a Tensor
        sr = sr.item()  # Extract the value as a Python number

    # 2. Convert y_true and y_pred to NumPy arrays
    y_true_np = y_true.cpu().detach().numpy()[0]  # .cpu() is crucial!
    y_pred_np = y_pred.cpu().detach().numpy()[0]


    mfccs_true = librosa.feature.mfcc(y=y_true_np, sr=sr, n_mfcc=20)
    mfccs_pred = librosa.feature.mfcc(y=y_pred_np, sr=sr, n_mfcc=20)

    # 3. Convert MFCCs back to PyTorch tensors and ensure correct device
    mfccs_true = torch.tensor(mfccs_true, device=y_true.device, dtype=torch.float32)
    mfccs_pred = torch.tensor(mfccs_pred, device=y_pred.device, dtype=torch.float32)

    return torch.mean((mfccs_true - mfccs_pred)**2)

def discriminator_train(high_quality, low_quality, real_labels, fake_labels):
    optimizer_d.zero_grad()

    # Forward pass for real samples
    discriminator_decision_from_real = discriminator(high_quality[0])
    d_loss_real = criterion_d(discriminator_decision_from_real, real_labels)

    # Forward pass for fake samples (from generator output)
    generator_output = generator(low_quality[0])
    discriminator_decision_from_fake = discriminator(generator_output.detach())
    d_loss_fake = criterion_d(discriminator_decision_from_fake, fake_labels)

    # Combine real and fake losses
    d_loss = (d_loss_real + d_loss_fake) / 2.0

    # Backward pass and optimization
    d_loss.backward()
    nn.utils.clip_grad_norm_(discriminator.parameters(), max_norm=1.0)  # Gradient Clipping
    optimizer_d.step()

    return d_loss

def generator_train(low_quality, high_quality, real_labels):
    optimizer_g.zero_grad()

    # Forward pass for fake samples (from generator output)
    generator_output = generator(low_quality[0])

    mfcc_l = mfcc_loss(high_quality[0], generator_output, high_quality[1])

    discriminator_decision = discriminator(generator_output)
    adversarial_loss = criterion_g(discriminator_decision, real_labels)

    combined_loss = adversarial_loss + 0.5 * mfcc_l

    combined_loss.backward()
    optimizer_g.step()

    return (generator_output, combined_loss, adversarial_loss, mfcc_l)

# Init script argument parser
parser = argparse.ArgumentParser(description="Training script")
parser.add_argument("--generator", type=str, default=None,
                    help="Path to the generator model file")
parser.add_argument("--discriminator", type=str, default=None,
                    help="Path to the discriminator model file")
parser.add_argument("--verbose", action="store_true",  help="Increase output verbosity")

args = parser.parse_args()

# Check for CUDA availability
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
print(f"Using device: {device}")

debug = args.verbose

# Initialize dataset and dataloader
dataset_dir = './dataset/good'
dataset = AudioDataset(dataset_dir)

# ========= MULTIPLE =========

# dataset_size = len(dataset)
# train_size = int(dataset_size * .9)
# val_size = int(dataset_size-train_size)

#train_dataset, val_dataset = random_split(dataset, [train_size, val_size])

# train_data_loader = DataLoader(train_dataset, batch_size=1, shuffle=True)
# val_data_loader = DataLoader(val_dataset, batch_size=1, shuffle=True)

# ========= SINGLE =========

train_data_loader = DataLoader(dataset, batch_size=1, shuffle=True)

# Initialize models and move them to device
generator = SISUGenerator()
discriminator = SISUDiscriminator()

if args.generator is not None:
    generator.load_state_dict(torch.load(args.generator, weights_only=True))
if args.discriminator is not None:
    discriminator.load_state_dict(torch.load(args.discriminator, weights_only=True))

generator = generator.to(device)
discriminator = discriminator.to(device)

# Loss
criterion_g = nn.MSELoss()
criterion_d = nn.BCELoss()

# Optimizers
optimizer_g = optim.Adam(generator.parameters(), lr=0.0001, betas=(0.5, 0.999))
optimizer_d = optim.Adam(discriminator.parameters(), lr=0.0001, betas=(0.5, 0.999))

# Scheduler
scheduler_g = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer_g, mode='min', factor=0.5, patience=5)
scheduler_d = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer_d, mode='min', factor=0.5, patience=5)

models_dir = "models"
os.makedirs(models_dir, exist_ok=True)

def start_training():
    generator_epochs = 5000
    for generator_epoch in range(generator_epochs):
        low_quality_audio = (torch.empty((1)), 1)
        high_quality_audio = (torch.empty((1)), 1)
        ai_enhanced_audio = (torch.empty((1)), 1)

        times_correct = 0

        # ========= TRAINING =========
        for high_quality_clip, low_quality_clip in tqdm.tqdm(train_data_loader, desc=f"Epoch {generator_epoch+1}/{generator_epochs}"):
        # for high_quality_clip, low_quality_clip in train_data_loader:
            high_quality_sample = (high_quality_clip[0].to(device), high_quality_clip[1])
            low_quality_sample = (low_quality_clip[0].to(device), low_quality_clip[1])

            # ========= LABELS =========
            batch_size = high_quality_clip[0].size(0)
            real_labels = torch.ones(batch_size, 1).to(device)
            fake_labels = torch.zeros(batch_size, 1).to(device)

            # ========= DISCRIMINATOR =========
            discriminator.train()
            d_loss = discriminator_train(high_quality_sample, low_quality_sample, real_labels, fake_labels)

            # ========= GENERATOR =========
            generator.train()
            generator_output, combined_loss, adversarial_loss, mfcc_l = generator_train(low_quality_sample, high_quality_sample, real_labels)

            if debug:
                print(d_loss, combined_loss, adversarial_loss, mfcc_l)
            scheduler_d.step(d_loss)
            scheduler_g.step(combined_loss)

            # ========= SAVE LATEST AUDIO =========
            high_quality_audio = high_quality_clip
            low_quality_audio = low_quality_clip
            ai_enhanced_audio = (generator_output, high_quality_clip[1])

        if generator_epoch % 10 == 0:
            print(f"Saved epoch {generator_epoch}!")
            torchaudio.save(f"./output/epoch-{generator_epoch}-audio-crap.wav", low_quality_audio[0][0].cpu(), high_quality_audio[1]) # <-- Because audio clip was resampled in data.py from original to crap and to original again.
            torchaudio.save(f"./output/epoch-{generator_epoch}-audio-ai.wav", ai_enhanced_audio[0][0].cpu(), ai_enhanced_audio[1])
            torchaudio.save(f"./output/epoch-{generator_epoch}-audio-orig.wav", high_quality_audio[0][0].cpu(), high_quality_audio[1])

        torch.save(discriminator.state_dict(), f"{models_dir}/discriminator_epoch_{generator_epoch}.pt")
        torch.save(generator.state_dict(), f"{models_dir}/generator_epoch_{generator_epoch}.pt")
        torch.save(discriminator, f"{models_dir}/discriminator_epoch_{generator_epoch}_full.pt")
        torch.save(generator, f"{models_dir}/generator_epoch_{generator_epoch}_full.pt")

    torch.save(discriminator, "models/epoch-5000-discriminator.pt")
    torch.save(generator, "models/epoch-5000-generator.pt")
    print("Training complete!")

start_training()
